#define BAS_PRIVATE 1

#include "BetterAuthorizationSampleLib.h"

#include <launch.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <CoreServices/CoreServices.h>

enum {
    kIdleTimeoutInSeconds     = 120,
    kWatchdogTimeoutInSeconds = 65
};

#define kBASMaxNumberOfKBytes			(1024 * 1024)
#define kBASSocketPathFormat			"/var/run/%s.socket"
#define kLaunchDSocketDictKey           "MasterSocket"

#pragma mark ***** Common Code

extern int BASOSStatusToErrno(OSStatus errNum)
    // See comment in header.
{
    int retval;

    #define CASE(ident)         \
        case k ## ident ## Err: \
            retval = ident;     \
            break
    switch (errNum) {
        case noErr:
            retval = 0;
            break;
        case kENORSRCErr:
            retval = ESRCH;                 // no ENORSRC on Mac OS X, so use ESRCH
            break;
        case memFullErr:
            retval = ENOMEM;
            break;
        CASE(EDEADLK);
        CASE(EAGAIN);
        case kEOPNOTSUPPErr:
            retval = ENOTSUP;
            break;
        CASE(EPROTO);
        CASE(ETIME);
        CASE(ENOSR);
        CASE(EBADMSG);
        case kECANCELErr:
            retval = ECANCELED;             // note spelling difference
            break;
        CASE(ENOSTR);
        CASE(ENODATA);
        CASE(EINPROGRESS);
        CASE(ESRCH);
        CASE(ENOMSG);
        default:
            if ( (errNum <= kEPERMErr) && (errNum >= kENOMSGErr) ) {
                retval = (-3200 - errNum) + 1;				// OT based error
            } else if ( (errNum >= errSecErrnoBase) && (errNum <= (errSecErrnoBase + ELAST)) ) {
                retval = (int) errNum - errSecErrnoBase;	// POSIX based error
            } else {
                retval = (int) errNum;						// just return the value unmodified
            }
    }
    #undef CASE
    return retval;
}

extern OSStatus BASErrnoToOSStatus(int errNum)
    // See comment in header.
{
    OSStatus retval;
    
    if ( errNum == 0 ) {
        retval = noErr;
    } else if ( (errNum >= EPERM) && (errNum <= ELAST) ) {
        retval = (OSStatus) errNum + errSecErrnoBase;
    } else {
        retval = (int) errNum;      // just return the value unmodified
    }
    
    return retval;
}

static Boolean BASIsBinaryPropertyListData(const void * plistBuffer, size_t plistSize)
{
    static const char kBASBinaryPlistWatermark[6] = "bplist";
    
    assert(plistBuffer != NULL);
    
    return (plistSize >= sizeof(kBASBinaryPlistWatermark)) 
        && (memcmp(plistBuffer, kBASBinaryPlistWatermark, sizeof(kBASBinaryPlistWatermark)) == 0);
}

static void NormaliseOSStatusErrorCode(OSStatus *errPtr)
{
    assert(errPtr != NULL);
    
    if ( (*errPtr == errAuthorizationCanceled) || (*errPtr == (errSecErrnoBase + ECANCELED)) ) {
        *errPtr = userCanceledErr;
    }
}

static int BASRead(int fd, void *buf, size_t bufSize, size_t *bytesRead)
{
    int 	err;
    char *	cursor;
    size_t	bytesLeft;
    ssize_t bytesThisTime;

    assert(fd >= 0);
    assert(buf != NULL);
    // bufSize may be 0
    assert(bufSize <= kBASMaxNumberOfKBytes);

    err = 0;
    bytesLeft = bufSize;
    cursor = (char *) buf;
    while ( (err == 0) && (bytesLeft != 0) ) {
        bytesThisTime = read(fd, cursor, bytesLeft);
        if (bytesThisTime > 0) {
            cursor    += bytesThisTime;
            bytesLeft -= bytesThisTime;
        } else if (bytesThisTime == 0) {
            err = EPIPE;
        } else {
            assert(bytesThisTime == -1);
            
            err = errno;
            assert(err != 0);
            if (err == EINTR) {
                err = 0;		// let's loop again
            }
        }
    }
    if (bytesRead != NULL) {
        *bytesRead = bufSize - bytesLeft;
    }

    return err;
}

static int BASWrite(int fd, const void *buf, size_t bufSize, size_t *bytesWritten)
{
    int 	err;
    char *	cursor;
    size_t	bytesLeft;
    ssize_t bytesThisTime;

    assert(fd >= 0);
    assert(buf != NULL);
    // bufSize may be 0
    assert(bufSize <= kBASMaxNumberOfKBytes);

    #if !defined(BAS_WRITE_CHECK_SIGPIPE)
        #define BAS_WRITE_CHECK_SIGPIPE 1
    #endif
    #if !defined(NDEBUG) && BAS_WRITE_CHECK_SIGPIPE
        {
            int					junk;
            struct stat			sb;
            struct sigaction	currentSignalState;
            int					val;
            socklen_t			valLen;
            
            junk = fstat(fd, &sb);
            assert(junk == 0);

            if ( S_ISFIFO(sb.st_mode) || S_ISSOCK(sb.st_mode) ) {
                junk = sigaction(SIGPIPE, NULL, &currentSignalState);
                assert(junk == 0);
                
                valLen = sizeof(val);
                junk = getsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, &val, &valLen);
                assert(junk == 0);
                assert(valLen == sizeof(val));

                assert( (currentSignalState.sa_handler == SIG_IGN) || (val == 1) );
            }
        }
    #endif

    err = 0;
    bytesLeft = bufSize;
    cursor = (char *) buf;
    while ( (err == 0) && (bytesLeft != 0) ) {
        bytesThisTime = write(fd, cursor, bytesLeft);
        if (bytesThisTime > 0) {
            cursor    += bytesThisTime;
            bytesLeft -= bytesThisTime;
        } else if (bytesThisTime == 0) {
            assert(false);
            err = EPIPE;
        } else {
            assert(bytesThisTime == -1);

            err = errno;
            assert(err != 0);
            if (err == EINTR) {
                err = 0;		// let's loop again
            }
        }
    }
    if (bytesWritten != NULL) {
        *bytesWritten = bufSize - bytesLeft;
    }

    return err;
}

static int BASReadDictionary(int fdIn, CFDictionaryRef *dictPtr)
{
    int                 err = 0;
    uint32_t			dictSize;
    void *				dictBuffer;
    CFDataRef			dictData;
    CFPropertyListRef 	dict;

    assert(fdIn >= 0);
    assert( dictPtr != NULL);
    assert(*dictPtr == NULL);
    
    dictBuffer = NULL;
    dictData   = NULL;
    dict       = NULL;

    err = BASRead(fdIn, &dictSize, sizeof(dictSize), NULL);
    if (err == 0) {
        dictSize = OSSwapBigToHostInt32(dictSize);
        if (dictSize == 0) {
            err = EINVAL;
        } else if (dictSize > kBASMaxNumberOfKBytes) {
            err = EINVAL;
        }
    }
    if (err == 0) {
        dictBuffer = malloc( (size_t) dictSize);
        if (dictBuffer == NULL) {
            err = ENOMEM;
        }
    }

    if (err == 0) {
        err = BASRead(fdIn, dictBuffer, dictSize, NULL);
    }
    if ( (err == 0) && BASIsBinaryPropertyListData(dictBuffer, dictSize) ) {
        err = BASOSStatusToErrno( coreFoundationUnknownErr );
    }
    if (err == 0) {
        dictData = CFDataCreateWithBytesNoCopy(NULL, dictBuffer, dictSize, kCFAllocatorNull);
        if (dictData == NULL) {
            err = BASOSStatusToErrno( coreFoundationUnknownErr );
        }
    }
    if (err == 0) {
        dict = CFPropertyListCreateFromXMLData(NULL, dictData, kCFPropertyListImmutable, NULL);
        if (dict == NULL) {
            err = BASOSStatusToErrno( coreFoundationUnknownErr );
        }
    }
    if ( (err == 0) && (CFGetTypeID(dict) != CFDictionaryGetTypeID()) ) {
        err = EINVAL;		// only CFDictionaries need apply
    }

    if (err != 0) {
        if (dict != NULL) {
            CFRelease(dict);
        }
        dict = NULL;
    }
    *dictPtr = (CFDictionaryRef) dict;
    free(dictBuffer);
    if (dictData != NULL) {
        CFRelease(dictData);
    }

    assert( (err == 0) == (*dictPtr != NULL) );

    return err;
}

static int BASWriteDictionary(CFDictionaryRef dict, int fdOut)
{
    int                 err = 0;
    CFDataRef			dictData;
    uint32_t			dictSize;

    assert(dict != NULL);
    assert(fdOut >= 0);
    
    dictData   = NULL;

    dictData = CFPropertyListCreateXMLData(NULL, dict);
    if (dictData == NULL) {
        err = BASOSStatusToErrno( coreFoundationUnknownErr );
    }

    if ( (err == 0) && (CFDataGetLength(dictData) > kBASMaxNumberOfKBytes) ) {
        err = EINVAL;
    }
    if (err == 0) {
        dictSize = OSSwapHostToBigInt32( CFDataGetLength(dictData) );
        err = BASWrite(fdOut, &dictSize, sizeof(dictSize), NULL);
    }
    if (err == 0) {
        err = BASWrite(fdOut, CFDataGetBytePtr(dictData), CFDataGetLength(dictData), NULL);
    }

    if (dictData != NULL) {
        CFRelease(dictData);
    }
    return err;
}

static const char kDummyData = 'D';

static const char kACKData   = 'A';

static int BASReadDescriptor(int fd, int *fdRead)
{
    int 				err;
    int 				junk;
    struct msghdr 		msg;
    struct iovec		iov;
    struct {
        struct cmsghdr 	hdr;
        int            	fd;
    } 					control;
    char				dummyData;
    ssize_t				bytesReceived;

    assert(fd >= 0);
    assert( fdRead != NULL);
    assert(*fdRead == -1);

    iov.iov_base = (char *) &dummyData;
    iov.iov_len  = sizeof(dummyData);

    msg.msg_name       = NULL;
    msg.msg_namelen    = 0;
    msg.msg_iov        = &iov;
    msg.msg_iovlen     = 1;
    msg.msg_control    = (caddr_t) &control;
    msg.msg_controllen = sizeof(control);
    msg.msg_flags	   = MSG_WAITALL;

    do {
        bytesReceived = recvmsg(fd, &msg, 0);
        if (bytesReceived == sizeof(dummyData)) {
            if (   (dummyData != kDummyData)
                || (msg.msg_flags != 0) 
                || (msg.msg_control == NULL) 
                || (msg.msg_controllen != sizeof(control)) 
                || (control.hdr.cmsg_len != sizeof(control)) 
                || (control.hdr.cmsg_level != SOL_SOCKET)
                || (control.hdr.cmsg_type  != SCM_RIGHTS) 
                || (control.fd < 0) ) {
                err = EINVAL;
            } else {
                *fdRead = control.fd;
                err = 0;
            }
        } else if (bytesReceived == 0) {
            err = EPIPE;
        } else {
            assert(bytesReceived == -1);

            err = errno;
            assert(err != 0);
        }
    } while (err == EINTR);

    if (err == 0) {
        err = BASWrite(fd, &kACKData, sizeof(kACKData), NULL);
        if (err != 0) {
            junk = close(*fdRead);
            assert(junk == 0);
            *fdRead = -1;
        }
    }

    assert( (err == 0) == (*fdRead >= 0) );

    return err;
}

static int BASWriteDescriptor(int fd, int fdToWrite)
{
    int 				err;
    struct msghdr 		msg;
    struct iovec		iov;
    struct {
        struct cmsghdr 	hdr;
        int            	fd;
    } 					control;
    ssize_t 			bytesSent;
    char                ack;

    assert(fd >= 0);
    assert(fdToWrite >= 0);

    control.hdr.cmsg_len   = sizeof(control);
    control.hdr.cmsg_level = SOL_SOCKET;
    control.hdr.cmsg_type  = SCM_RIGHTS;
    control.fd             = fdToWrite;

    iov.iov_base = (char *) &kDummyData;
    iov.iov_len  = sizeof(kDummyData);

    msg.msg_name       = NULL;
    msg.msg_namelen    = 0;
    msg.msg_iov        = &iov;
    msg.msg_iovlen     = 1;
    msg.msg_control    = (caddr_t) &control;
    msg.msg_controllen = control.hdr.cmsg_len;
    msg.msg_flags	   = 0;
    do {
        bytesSent = sendmsg(fd, &msg, 0);
        if (bytesSent == sizeof(kDummyData)) {
            err = 0;
        } else {
            assert(bytesSent == -1);

            err = errno;
            assert(err != 0);
        }
    } while (err == EINTR);

    if (err == 0) {
        err = BASRead(fd, &ack, sizeof(ack), NULL);
        if ( (err == 0) && (ack != kACKData) ) {
            err = EINVAL;
        }
    }

    return err;
}

extern void BASCloseDescriptorArray(
    CFArrayRef					descArray
)
{
    int							junk;
    CFIndex						descCount;
    CFIndex						descIndex;

    if (descArray != NULL) {
        if (CFGetTypeID(descArray) == CFArrayGetTypeID()) {
            descCount = CFArrayGetCount(descArray);

            for (descIndex = 0; descIndex < descCount; descIndex++) {
                CFNumberRef thisDescNum;
                int 		thisDesc;
        
                thisDescNum = (CFNumberRef) CFArrayGetValueAtIndex(descArray, descIndex);
                if (   (thisDescNum == NULL) 
                    || (CFGetTypeID(thisDescNum) != CFNumberGetTypeID()) 
                    || ! CFNumberGetValue(thisDescNum, kCFNumberIntType, &thisDesc) ) {
                    assert(false);
                } else {
                    assert(thisDesc >= 0);
                    junk = close(thisDesc);
                    assert(junk == 0);
                }
            }
        } else {
            assert(false);
        }
    }
}

static int BASReadDictioanaryTranslatingDescriptors(int fd, CFDictionaryRef *dictPtr)
{
    int 				err;
    int 				junk;
    CFDictionaryRef		dict;
    CFArrayRef 			incomingDescs;

    assert(fd >= 0);
    assert( dictPtr != NULL);
    assert(*dictPtr == NULL);

    dict = NULL;

    err = BASReadDictionary(fd, &dict);

    if (err == 0) {
        incomingDescs = (CFArrayRef) CFDictionaryGetValue(dict, CFSTR(kBASDescriptorArrayKey));
        if (incomingDescs == NULL) {
            *dictPtr = dict;
            dict = NULL;
        } else {
            CFMutableArrayRef 		translatedDescs;
            CFMutableDictionaryRef	mutableDict;
            CFIndex					descCount;
            CFIndex					descIndex;

            translatedDescs  = NULL;
            mutableDict      = NULL;

            if ( CFGetTypeID(incomingDescs) != CFArrayGetTypeID() ) {
                err = EINVAL;
            }

            if (err == 0) {
                translatedDescs = CFArrayCreateMutable(NULL, 0, &kCFTypeArrayCallBacks);
                if (translatedDescs == NULL) {
                    err = coreFoundationUnknownErr;
                }
            }
            if (err == 0) {
                mutableDict = CFDictionaryCreateMutableCopy(NULL, 0, dict);
                if (mutableDict == NULL) {
                    err = BASOSStatusToErrno( coreFoundationUnknownErr );
                }
            }

            if (err == 0) {
                descCount = CFArrayGetCount(incomingDescs);

                #if !defined(NDEBUG)
                    for (descIndex = 0; descIndex < descCount; descIndex++) {
                        int 		thisDesc;
                        CFNumberRef thisDescNum;
                        
                        thisDescNum = (CFNumberRef) CFArrayGetValueAtIndex(incomingDescs, descIndex);
                        assert(thisDescNum != NULL);
                        assert(CFGetTypeID(thisDescNum) == CFNumberGetTypeID());
                        assert(CFNumberGetValue(thisDescNum, kCFNumberIntType, &thisDesc));
                        assert(thisDesc >= 0);
                    }
                #endif

                for (descIndex = 0; descIndex < descCount; descIndex++) {
                    int 		thisDesc;
                    CFNumberRef thisDescNum;

                    thisDesc = -1;
                    thisDescNum = NULL;

                    err = BASReadDescriptor(fd, &thisDesc);
                    if (err == 0) {
                        thisDescNum = CFNumberCreate(NULL, kCFNumberIntType, &thisDesc);
                        if (thisDescNum == NULL) {
                            err = BASOSStatusToErrno( coreFoundationUnknownErr );
                        }
                    }
                    if (err == 0) {
                        CFArrayAppendValue(translatedDescs, thisDescNum);

                        thisDesc = -1;		
                    }
                    
                    if (thisDescNum != NULL) {
                        CFRelease(thisDescNum);
                    }
                    if (thisDesc != -1) {
                        junk = close(thisDesc);
                        assert(junk == 0);
                    }
                    
                    if (err != 0) {
                        break;
                    }
                }
            }

            if (err == 0) {
                CFDictionarySetValue(mutableDict, CFSTR(kBASDescriptorArrayKey), translatedDescs);
                *dictPtr = mutableDict;
            } else {
                BASCloseDescriptorArray(translatedDescs);
                if (mutableDict != NULL) {
                    CFRelease(mutableDict);
                }
            }
            if (translatedDescs != NULL) {
                CFRelease(translatedDescs);
            }
        }
    }
    
    if (dict != NULL) {
        CFRelease(dict);
    }

    assert( (err == 0) == (*dictPtr != NULL) );

    return err;
}

static OSStatus FindCommand(
    CFDictionaryRef             request,
    const BASCommandSpec		commands[],
    size_t *                    commandIndexPtr
)
{
    OSStatus					retval = noErr;
    CFStringRef                 commandStr;
    char *                      command;
    long						commandSize = 0;
    size_t						index = 0;

    assert(request != NULL);
    assert(commands != NULL);
    assert(commands[0].commandName != NULL);        // there must be at least one command
    assert(commandIndexPtr != NULL);

    command = NULL;

    commandStr = CFDictionaryGetValue(request, CFSTR(kBASCommandKey));
    if ( (commandStr == NULL) || (CFGetTypeID(commandStr) != CFStringGetTypeID()) ) {
        retval = paramErr;
    }
    commandSize = CFStringGetLength(commandStr);
    if ( (retval == noErr) && (commandSize > 1024) ) {
        retval = paramErr;
    }
    if (retval == noErr) {
        size_t      bufSize;

        bufSize = CFStringGetMaximumSizeForEncoding(CFStringGetLength(commandStr), kCFStringEncodingUTF8) + 1;
        command = malloc(bufSize);

        if (command == NULL) {
            retval = memFullErr;
        } else if ( ! CFStringGetCString(commandStr, command, bufSize, kCFStringEncodingUTF8) ) {
            retval = coreFoundationUnknownErr;
        }
    }

    if (retval == noErr) {
        do {
            if ( strcmp(commands[index].commandName, command) == 0 ) {
                *commandIndexPtr = index;
                break;
            }
            index += 1;
            if (commands[index].commandName == NULL) {
                retval = BASErrnoToOSStatus(ENOENT);
                break;
            }
        } while (true);
    }

    free(command);

    return retval;
}

extern OSStatus BASExecuteRequestInHelperTool(
    AuthorizationRef			auth,
    const BASCommandSpec		commands[],
    CFStringRef					bundleID,
    CFDictionaryRef				request,
    CFDictionaryRef *			response
)
{	
    OSStatus					retval = noErr;
    int                         junk;
    size_t                      commandIndex;
    char                        bundleIDC[PATH_MAX];
    int							fd = -1;
    struct sockaddr_un			addr;
    AuthorizationExternalForm	extAuth;

    assert(auth != NULL);
    assert(commands != NULL);
    assert(commands[0].commandName != NULL);
    assert(bundleID != NULL);
    assert(request != NULL);
    assert( response != NULL);
    assert(*response == NULL);

    assert(CFDictionaryContainsKey(request, CFSTR(kBASCommandKey)));
    assert(CFGetTypeID(CFDictionaryGetValue(request, CFSTR(kBASCommandKey))) == CFStringGetTypeID());

    retval = FindCommand(request, commands, &commandIndex);

    #if !defined(BAS_PREAUTHORIZE)
        #define BAS_PREAUTHORIZE 1
    #endif
    #if BAS_PREAUTHORIZE
        if ( (retval == noErr) && (commands[commandIndex].rightName != NULL) ) {
            AuthorizationItem   item   = { commands[commandIndex].rightName, 0, NULL, 0 };
            AuthorizationRights rights = { 1, &item };

            retval = AuthorizationCopyRights(auth, &rights, kAuthorizationEmptyEnvironment, kAuthorizationFlagExtendRights | kAuthorizationFlagInteractionAllowed | kAuthorizationFlagPreAuthorize, NULL);
        }
    #endif

    if (retval == noErr) {
        fd = socket(AF_UNIX, SOCK_STREAM, 0);
        if (fd == -1) { 
            retval = BASErrnoToOSStatus(errno);
        }
    }
    if (retval == noErr) {
        static const int kOne = 1;
        
        if ( setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, &kOne, sizeof(kOne)) < 0 ) {
            retval = BASErrnoToOSStatus(errno);
        }
    }

    if (retval == noErr) {
        if ( ! CFStringGetFileSystemRepresentation(bundleID, bundleIDC, sizeof(bundleIDC)) ) {
            retval = coreFoundationUnknownErr;
        }
    }
    if (retval == noErr) {
        int         pathLen;

        memset(&addr, 0, sizeof(addr));

        addr.sun_family = AF_UNIX;
        pathLen = snprintf(addr.sun_path, sizeof(addr.sun_path), kBASSocketPathFormat, bundleIDC);
        if (pathLen >= (int)sizeof(addr.sun_path)) {
            retval = paramErr;
        } else {
            addr.sun_len = SUN_LEN(&addr);
        }
    }

    if (retval == noErr) {
        if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
            retval = BASErrnoToOSStatus(errno);
        }
    }

    if (retval == noErr) {
        retval = AuthorizationMakeExternalForm(auth, &extAuth);
    }
    if (retval == noErr) {	
        retval = BASErrnoToOSStatus( BASWrite(fd, &extAuth, sizeof(extAuth), NULL) );
    }

    if (retval == noErr) {	
        retval = BASErrnoToOSStatus( BASWriteDictionary(request, fd) );
    }

    if (retval == noErr) {
        retval = BASErrnoToOSStatus( BASReadDictioanaryTranslatingDescriptors(fd, response) );
    }

    if (fd != -1) {
        junk = close(fd);
        assert(junk == 0);
    }
    NormaliseOSStatusErrorCode(&retval);

    assert( (retval == noErr) == (*response != NULL) );

    return retval;
}